# type: ignore

import os
from unittest.mock import mock_open

import pytest
import pytest_mock

# Public API
from dbos import DBOS, load_config
from dbos._dbos_config import (
    ConfigFile,
    DBOSConfig,
    check_config_consistency,
    overwrite_config,
    parse_database_url_to_dbconfig,
    process_config,
    set_env_vars,
    translate_dbos_config_to_config_file,
)
from dbos._error import DBOSInitializationError

mock_filename = "dbos-config.yaml"
original_open = __builtins__["open"]


def generate_mock_open(filenames, mock_files):
    if not isinstance(filenames, list):
        filenames = [filenames]
    if not isinstance(mock_files, list):
        mock_files = [mock_files]

    def conditional_mock_open(*args, **kwargs):
        for filename, mock_file in zip(filenames, mock_files):
            if args[0] == filename:
                m = mock_open(read_data=mock_file)
                return m()
        return original_open(*args, **kwargs)

    return conditional_mock_open


"""
Test all the possible ways to configure DBOS.
- First test the "switches" in the DBOS.__init__() method, ensuring they find the config from the right place
- Then test each of the individual functions that process the config (load_config, overwrite_config, translate_dbos_config_to_config_file, process_config)
"""


####################
# SWITCHES
####################
def test_no_config_provided(mocker):
    mock_config = """
        name: "some-app"
        language: "python"
        runtimeConfig:
            start:
                - "python3 main.py"
            admin_port: 8001
        database:
          hostname: 'localhost'
          port: 5432
          username: 'postgres'
          password: ${PGPASSWORD}
          app_db_name: 'some db'
          connectionTimeoutMillis: 3000
        env:
            foo: ${BARBAR}
            bazbaz: BAZBAZ
            bob: ${BOBBOB}
            test_number: 123
    """
    os.environ["BARBAR"] = "FOOFOO"
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open(mock_filename, mock_config)
    )

    dbos = DBOS()
    assert dbos.config["name"] == "some-app"
    assert dbos.config["language"] == "python"
    assert dbos.config["database"]["hostname"] == "localhost"
    assert dbos.config["database"]["port"] == 5432
    assert dbos.config["database"]["username"] == "postgres"
    assert dbos.config["database"]["password"] == os.environ["PGPASSWORD"]
    assert dbos.config["database"]["app_db_name"] == "some db"
    assert dbos.config["database"]["connectionTimeoutMillis"] == 3000
    assert dbos.config["env"]["foo"] == "FOOFOO"
    assert dbos.config["env"]["bob"] is None  # Unset environment variable
    assert dbos.config["env"]["test_number"] == 123

    set_env_vars(dbos.config)
    assert os.environ["bazbaz"] == "BAZBAZ"
    assert os.environ["foo"] == "FOOFOO"
    assert os.environ["test_number"] == "123"
    assert "bob" not in os.environ

    dbos.destroy()


def test_configfile_type_provided():
    config: ConfigFile = {
        "name": "some-app",
        "database": {
            "hostname": "localhost",
        },
    }
    dbos = DBOS(config=config)
    assert dbos.config["name"] == "some-app"
    assert dbos.config["database"]["hostname"] == "localhost"
    assert dbos.config["database"]["port"] == 5432
    assert dbos.config["database"]["username"] == "postgres"
    assert dbos.config["database"]["password"] == os.environ["PGPASSWORD"]
    assert dbos.config["database"]["app_db_name"] == "some_app"
    dbos.destroy()


def test_dbosconfig_type_provided(mocker):
    config: DBOSConfig = {
        "name": "some-app",
    }
    dbos = DBOS(config=config)
    assert dbos.config["name"] == "some-app"
    assert dbos.config["database"]["hostname"] == "localhost"
    assert dbos.config["database"]["port"] == 5432
    assert dbos.config["database"]["username"] == "postgres"
    assert dbos.config["database"]["password"] == os.environ["PGPASSWORD"]
    assert dbos.config["database"]["app_db_name"] == "some_app"
    dbos.destroy()


####################
# LOAD CONFIG
####################


def test_load_valid_config_file(mocker):
    mock_config = """
        name: "some-app"
        runtimeConfig:
            start:
                - "python3 main.py"
            admin_port: 8001
        database:
          hostname: 'localhost'
          port: 5432
          username: 'postgres'
          password: ${PGPASSWORD}
          app_db_name: 'some db'
          connectionTimeoutMillis: 3000
        env:
            foo: ${BARBAR}
            bazbaz: BAZBAZ
            bob: ${BOBBOB}
            test_number: 123
    """
    os.environ["BARBAR"] = "FOOFOO"
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open(mock_filename, mock_config)
    )

    configFile = load_config(mock_filename, run_process_config=False)
    assert configFile["name"] == "some-app"
    assert configFile["database"]["hostname"] == "localhost"
    assert configFile["database"]["port"] == 5432
    assert configFile["database"]["username"] == "postgres"
    assert configFile["database"]["password"] == os.environ["PGPASSWORD"]
    assert configFile["database"]["app_db_name"] == "some db"
    assert configFile["database"]["connectionTimeoutMillis"] == 3000
    assert "database_url" not in configFile
    assert configFile["env"]["foo"] == "FOOFOO"
    assert configFile["env"]["bob"] is None  # Unset environment variable
    assert configFile["env"]["test_number"] == 123

    set_env_vars(configFile)
    assert os.environ["bazbaz"] == "BAZBAZ"
    assert os.environ["foo"] == "FOOFOO"
    assert os.environ["test_number"] == "123"
    assert "bob" not in os.environ


def test_load_config_database_url(mocker):
    mock_config = """
        name: "some-app"
        database_url: "postgresql://user:password@localhost:5432/dbname?connect_timeout=10&sslmode=require&sslrootcert=ca.pem"
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open(mock_filename, mock_config)
    )

    configFile = load_config(mock_filename, run_process_config=False)
    assert configFile["name"] == "some-app"
    assert (
        configFile["database_url"]
        == "postgresql://user:password@localhost:5432/dbname?connect_timeout=10&sslmode=require&sslrootcert=ca.pem"
    )
    assert "database" not in configFile


def test_load_config_database_url_and_database(mocker):
    mock_config = """
        name: "some-app"
        database_url: "postgresql://user:password@localhost:5432/dbname?connect_timeout=10&sslmode=require&sslrootcert=ca.pem"
        database:
            hostname: 'localhost'
            port: 5432
            username: 'postgres'
            password: ${PGPASSWORD}
            app_db_name: 'some db'
            connectionTimeoutMillis: 3000
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open(mock_filename, mock_config)
    )

    configFile = load_config(mock_filename, run_process_config=False)
    assert configFile["name"] == "some-app"
    assert (
        configFile["database_url"]
        == "postgresql://user:password@localhost:5432/dbname?connect_timeout=10&sslmode=require&sslrootcert=ca.pem"
    )
    assert configFile["database"]["hostname"] == "localhost"
    assert configFile["database"]["port"] == 5432
    assert configFile["database"]["username"] == "postgres"
    assert configFile["database"]["password"] == os.environ["PGPASSWORD"]
    assert configFile["database"]["app_db_name"] == "some db"
    assert configFile["database"]["connectionTimeoutMillis"] == 3000


def test_load_config_with_unset_database_url_env_var(mocker):
    mock_config = """
    name: "some-app"
    database_url: ${UNSET}
    """

    mocker.patch(
        "builtins.open", side_effect=generate_mock_open(mock_filename, mock_config)
    )

    configFile = load_config(mock_filename, run_process_config=False)
    assert configFile["name"] == "some-app"


def test_load_config_file_open_error(mocker):
    """Test handling when the config file can't be opened."""
    mocker.patch("builtins.open", side_effect=FileNotFoundError("File not found"))

    with pytest.raises(FileNotFoundError):
        load_config()


def test_load_config_file_not_a_dict(mocker):
    """Test handling when YAML doesn't parse to a dictionary."""
    mock_config = "just a string"
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    with pytest.raises(DBOSInitializationError) as exc_info:
        load_config()

    assert "must contain a dictionary" in str(exc_info.value)


def test_load_config_file_schema_validation_error(mocker):
    """Test handling when the config fails schema validation."""
    mock_config = """
    name: "test-app"
    invalid_field: "this shouldn't be here"
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    with pytest.raises(DBOSInitializationError) as exc_info:
        load_config()

    assert (
        "Validation error: Additional properties are not allowed ('invalid_field' was unexpected)"
        in str(exc_info.value)
    )


def test_load_config_file_custom_path():
    """Test parsing a config file from a custom path."""
    mock_config = """
    name: "test-app"
    database:
        hostname: "localhost"
    """
    custom_path = "/custom/path/dbos-config.yaml"
    from unittest.mock import mock_open, patch

    with patch("builtins.open", mock_open(read_data=mock_config)) as mock_file:
        result = load_config(custom_path, run_process_config=False)
        mock_file.assert_called_with(custom_path, "r")
        assert result["name"] == "test-app"


####################
# PROCESS CONFIG
####################


# Full config provided
def test_process_config_full():
    config: ConfigFile = {
        "name": "some-app",
        "database": {
            "hostname": "example.com",
            "port": 2345,
            "username": "example",
            "password": "password",
            "connectionTimeoutMillis": 3000,
            "app_db_name": "example_db",
            "sys_db_name": "sys_db",
            "ssl": True,
            "ssl_ca": "ca.pem",
            "local_suffix": False,
            "migrate": ["alembic upgrade head"],
            "rollback": ["alembic downgrade base"],
        },
        "runtimeConfig": {
            "start": ["python3 main.py"],
            "admin_port": 8001,
            "setup": ["echo 'hello'"],
        },
        "telemetry": {
            "logs": {
                "logLevel": "DEBUG",
            },
            "OTLPExporter": {
                "logsEndpoint": "thelogsendpoint",
                "tracesEndpoint": "thetracesendpoint",
            },
        },
        "env": {
            "FOO": "BAR",
        },
    }

    configFile = process_config(data=config, use_db_wizard=False)
    assert configFile["name"] == "some-app"
    assert configFile["database"]["hostname"] == "example.com"
    assert configFile["database"]["port"] == 2345
    assert configFile["database"]["username"] == "example"
    assert configFile["database"]["password"] == "password"
    assert configFile["database"]["connectionTimeoutMillis"] == 3000
    assert configFile["database"]["app_db_name"] == "example_db"
    assert configFile["database"]["sys_db_name"] == "sys_db"
    assert configFile["database"]["ssl"] == True
    assert configFile["database"]["ssl_ca"] == "ca.pem"
    assert configFile["database"]["local_suffix"] == False
    assert configFile["database"]["migrate"] == ["alembic upgrade head"]
    assert configFile["database"]["rollback"] == ["alembic downgrade base"]
    assert configFile["runtimeConfig"]["start"] == ["python3 main.py"]
    assert configFile["runtimeConfig"]["admin_port"] == 8001
    assert configFile["runtimeConfig"]["setup"] == ["echo 'hello'"]
    assert configFile["telemetry"]["logs"]["logLevel"] == "DEBUG"
    assert configFile["telemetry"]["OTLPExporter"]["logsEndpoint"] == "thelogsendpoint"
    assert (
        configFile["telemetry"]["OTLPExporter"]["tracesEndpoint"] == "thetracesendpoint"
    )
    assert configFile["env"]["FOO"] == "BAR"


def test_process_config_with_db_url():
    config: ConfigFile = {
        "name": "some-app",
        "database_url": "postgresql://user:password@localhost:7777/dbn?connect_timeout=1&sslmode=require&sslrootcert=ca.pem",
    }
    processed_config = process_config(data=config, use_db_wizard=False)
    assert processed_config["name"] == "some-app"
    assert processed_config["database"]["hostname"] == "localhost"
    assert processed_config["database"]["port"] == 7777
    assert processed_config["database"]["username"] == "user"
    assert processed_config["database"]["password"] == "password"
    assert processed_config["database"]["app_db_name"] == "dbn"
    assert processed_config["database"]["connectionTimeoutMillis"] == 1000
    assert processed_config["database"]["ssl"] == True
    assert processed_config["database"]["ssl_ca"] == "ca.pem"
    assert processed_config["database"]["local_suffix"] == False
    assert "rollback" not in processed_config["database"]
    assert "migrate" not in processed_config["database"]


def test_process_config_with_db_url_taking_precedence_over_database():
    config: ConfigFile = {
        "name": "some-app",
        "database": {
            "hostname": "example.com",
            "port": 2345,
            "username": "example",
            "password": "password",
            "connectionTimeoutMillis": 3000,
            "app_db_name": "example_db",
            "sys_db_name": "sys_db",
            "ssl": True,
            "ssl_ca": "ca.pem",
            "local_suffix": True,
            "migrate": ["alembic upgrade head"],
            "rollback": ["alembic downgrade base"],
        },
        "database_url": "postgresql://boo:whoisdiz@remotehost:7777/takesprecedence",
    }
    processed_config = process_config(data=config, use_db_wizard=False)
    assert processed_config["name"] == "some-app"
    assert processed_config["database"]["hostname"] == "remotehost"
    assert processed_config["database"]["port"] == 7777
    assert processed_config["database"]["username"] == "boo"
    assert processed_config["database"]["password"] == "whoisdiz"
    assert processed_config["database"]["app_db_name"] == "takesprecedence_local"
    assert processed_config["database"]["migrate"] == ["alembic upgrade head"]
    assert processed_config["database"]["rollback"] == ["alembic downgrade base"]
    assert processed_config["database"]["local_suffix"] == True
    assert "connectionTimeoutMillis" not in processed_config["database"]
    assert "ssl" not in processed_config["database"]
    assert "ssl_ca" not in processed_config["database"]


# Note this exercise going through the db wizard
def test_process_config_load_defaults():
    config: ConfigFile = {
        "name": "some-app",
    }
    processed_config = process_config(data=config)
    assert processed_config["name"] == "some-app"
    assert processed_config["database"]["app_db_name"] == "some_app"
    assert processed_config["database"]["hostname"] == "localhost"
    assert processed_config["database"]["port"] == 5432
    assert processed_config["database"]["username"] == "postgres"
    assert processed_config["database"]["password"] == os.environ.get(
        "PGPASSWORD", "dbos"
    )


def test_process_config_load_default_with_None_database_url():
    config: ConfigFile = {
        "name": "some-app",
        "database_url": None,
    }
    processed_config = process_config(data=config)
    assert processed_config["name"] == "some-app"
    assert processed_config["database"]["app_db_name"] == "some_app"
    assert processed_config["database"]["hostname"] == "localhost"
    assert processed_config["database"]["port"] == 5432
    assert processed_config["database"]["username"] == "postgres"
    assert processed_config["database"]["password"] == os.environ.get(
        "PGPASSWORD", "dbos"
    )


def test_process_config_with_None_app_db_name():
    config: ConfigFile = {
        "name": "some-app",
        "database": {
            "app_db_name": None,
        },
    }
    processed_config = process_config(data=config)
    print(processed_config)
    assert processed_config["name"] == "some-app"
    assert processed_config["database"]["app_db_name"] == "some_app"
    assert processed_config["database"]["hostname"] == "localhost"
    assert processed_config["database"]["port"] == 5432
    assert processed_config["database"]["username"] == "postgres"
    assert processed_config["database"]["password"] == os.environ.get(
        "PGPASSWORD", "dbos"
    )


def test_process_config_with_empty_app_db_name():
    config: ConfigFile = {
        "name": "some-app",
        "database": {
            "app_db_name": "",
        },
    }
    processed_config = process_config(data=config)
    print(processed_config)
    assert processed_config["name"] == "some-app"
    assert processed_config["database"]["app_db_name"] == "some_app"
    assert processed_config["database"]["hostname"] == "localhost"
    assert processed_config["database"]["port"] == 5432
    assert processed_config["database"]["username"] == "postgres"
    assert processed_config["database"]["password"] == os.environ.get(
        "PGPASSWORD", "dbos"
    )


def test_config_missing_name():
    config = {
        "database": {
            "hostname": "localhost",
            "port": 5432,
            "username": "postgres",
            "password": "dbos",
            "app_db_name": "dbostestpy",
        },
    }
    with pytest.raises(DBOSInitializationError) as exc_info:
        process_config(data=config)

    assert "must specify an application name" in str(exc_info.value)


def test_config_bad_name():
    config: ConfigFile = {
        "name": "some app",
    }
    with pytest.raises(DBOSInitializationError) as exc_info:
        process_config(data=config)
    assert "Invalid app name" in str(exc_info.value)


def test_load_config_load_db_connection(mocker):
    mock_db_connection = """
    {"hostname": "example.com", "port": 2345, "username": "example", "password": "password", "local_suffix": true}
    """
    mocker.patch(
        "builtins.open",
        side_effect=generate_mock_open([".dbos/db_connection"], [mock_db_connection]),
    )

    config = {
        "name": "some-app",
    }

    configFile = process_config(data=config, use_db_wizard=False)
    assert configFile["name"] == "some-app"
    assert configFile["database"]["hostname"] == "example.com"
    assert configFile["database"]["port"] == 2345
    assert configFile["database"]["username"] == "example"
    assert configFile["database"]["password"] == "password"
    assert configFile["database"]["local_suffix"] == True
    assert configFile["database"]["app_db_name"] == "some_app_local"


def test_config_mixed_params():
    config = {
        "name": "some-app",
        "database": {
            "port": 1234,
            "username": "some user",
            "password": "abc123",
        },
    }

    configFile = process_config(data=config, use_db_wizard=False)
    assert configFile["name"] == "some-app"
    assert configFile["database"]["hostname"] == "localhost"
    assert configFile["database"]["port"] == 1234
    assert configFile["database"]["username"] == "some user"
    assert configFile["database"]["password"] == "abc123"


def test_debug_override(mocker: pytest_mock.MockFixture):
    mocker.patch.dict(
        os.environ,
        {
            "DBOS_DBHOST": "fakehost",
            "DBOS_DBPORT": "1234",
            "DBOS_DBUSER": "fakeuser",
            "DBOS_DBPASSWORD": "fakepassword",
            "DBOS_DBLOCALSUFFIX": "false",
        },
    )

    config: Configfile = {
        "name": "some-app",
    }
    configFile = process_config(data=config, use_db_wizard=False)
    assert configFile["database"]["hostname"] == "fakehost"
    assert configFile["database"]["port"] == 1234
    assert configFile["database"]["username"] == "fakeuser"
    assert configFile["database"]["password"] == "fakepassword"
    assert configFile["database"]["local_suffix"] == False


def test_local_config():
    config: ConfigFile = {
        "name": "some-app",
        "database": {
            "hostname": "localhost",
            "port": 5432,
            "username": "postgres",
            "password": os.environ["PGPASSWORD"],
            "app_db_name": "some_db",
            "connectionTimeoutMillis": 3000,
            "local_suffix": True,
        },
    }
    processed_config = process_config(data=config)

    assert processed_config["name"] == "some-app"
    assert processed_config["database"]["local_suffix"] == True
    assert processed_config["database"]["hostname"] == "localhost"
    assert processed_config["database"]["port"] == 5432
    assert processed_config["database"]["username"] == "postgres"
    assert processed_config["database"]["password"] == os.environ["PGPASSWORD"]
    assert processed_config["database"]["app_db_name"] == "some_db_local"
    assert processed_config["database"]["connectionTimeoutMillis"] == 3000


def test_local_config_without_name(mocker):
    config: ConfigFile = {
        "name": "some-app",
        "database": {
            "hostname": "localhost",
            "port": 5432,
            "username": "postgres",
            "password": os.environ["PGPASSWORD"],
            "connectionTimeoutMillis": 3000,
            "local_suffix": True,
        },
    }
    processed_config = process_config(data=config)

    assert processed_config["name"] == "some-app"
    assert processed_config["database"]["local_suffix"] == True
    assert processed_config["database"]["hostname"] == "localhost"
    assert processed_config["database"]["port"] == 5432
    assert processed_config["database"]["username"] == "postgres"
    assert processed_config["database"]["password"] == os.environ["PGPASSWORD"]
    assert processed_config["database"]["app_db_name"] == "some_app_local"
    assert processed_config["database"]["connectionTimeoutMillis"] == 3000


####################
# DB STRING PARSING
####################


def test_basic_fields_mapping():
    """Test that basic fields from db_url are correctly mapped to db_config."""
    database_url = "postgresql://user:password@localhost:5432/dbname"
    db_config = parse_database_url_to_dbconfig(database_url)

    assert db_config["hostname"] == "localhost"
    assert db_config["port"] == 5432
    assert db_config["username"] == "user"
    assert db_config["password"] == "password"
    assert db_config["app_db_name"] == "dbname"


def test_no_db_name():
    """None app_db_name when dbname is not provided."""
    database_url = "postgresql://user:password@localhost:5432"
    db_config = parse_database_url_to_dbconfig(database_url)
    assert db_config["hostname"] == "localhost"
    assert db_config["port"] == 5432
    assert db_config["username"] == "user"
    assert db_config["password"] == "password"
    assert db_config["app_db_name"] == None


def test_no_db_name_end_with_slash():
    """Test that an exception is raised when dbname is not provided."""
    database_url = "postgresql://user:password@localhost:5432/"
    db_config = parse_database_url_to_dbconfig(database_url)
    assert db_config["hostname"] == "localhost"
    assert db_config["port"] == 5432
    assert db_config["username"] == "user"
    assert db_config["password"] == "password"
    assert db_config["app_db_name"] == ""


def test_default_port():
    """Test that default port (5432) is used when port is not specified."""
    database_url = "postgresql://user:password@localhost/dbname"
    db_config = parse_database_url_to_dbconfig(database_url)
    assert db_config["port"] == 5432


def test_query_parameters():
    """Test processing of various query parameters."""

    # Test connect_timeout conversion
    database_url = "postgresql://user:password@localhost:5432/dbname?connect_timeout=10"
    db_config = parse_database_url_to_dbconfig(database_url)
    assert db_config["connectionTimeoutMillis"] == 10000

    # Test sslmode=require (should set ssl=True)
    database_url = "postgresql://user:password@localhost:5432/dbname?sslmode=require"
    db_config = parse_database_url_to_dbconfig(database_url)
    assert db_config["ssl"] == True

    # Test sslmode=disable (should set ssl=False)
    database_url = "postgresql://user:password@localhost:5432/dbname?sslmode=disable"
    db_config = parse_database_url_to_dbconfig(database_url)
    assert db_config["ssl"] == False

    # Test sslmode=prefer (should set ssl=False as it's not 'require')
    database_url = "postgresql://user:password@localhost:5432/dbname?sslmode=prefer"
    db_config = parse_database_url_to_dbconfig(database_url)
    assert db_config["ssl"] == False

    # Test sslrootcert mapping to ssl_ca
    database_url = "postgresql://user:password@localhost:5432/dbname?sslrootcert=ca.pem"
    db_config = parse_database_url_to_dbconfig(database_url)
    assert db_config["ssl_ca"] == "ca.pem"

    # Test multiple parameters together
    database_url = "postgresql://user:password@localhost:5432/dbname?connect_timeout=10&sslmode=require&sslrootcert=ca.pem&application_name=myapp"
    db_config = parse_database_url_to_dbconfig(database_url)
    assert db_config["connectionTimeoutMillis"] == 10000
    assert db_config["ssl"] == True
    assert db_config["ssl_ca"] == "ca.pem"


def test_complex_password():
    """Test handling of complex passwords with special characters."""
    database_url = "postgresql://user:complex%23password@localhost:5432/dbname"
    db_config = parse_database_url_to_dbconfig(database_url)
    assert db_config["password"] == "complex#password"


def test_hostname_with_dots():
    """Test handling of hostnames with dots."""
    database_url = "postgresql://user:password@hostname.with.dots:5432/dbname"
    db_config = parse_database_url_to_dbconfig(database_url)
    assert db_config["hostname"] == "hostname.with.dots"


def test_invalid_string():
    """Test handling of invalid db strings."""
    with pytest.raises(Exception):
        database_url = "invalid"
        parse_database_url_to_dbconfig(database_url)


####################
# TRANSLATE DBOSConfig to ConfigFile
####################


def test_translate_dbosconfig_full_input():
    # Give all fields
    config: DBOSConfig = {
        "name": "test-app",
        "database_url": "postgresql://user:password@localhost:5432/dbname?connect_timeout=10&sslmode=require&sslrootcert=ca.pem",
        "sys_db_name": "sysdb",
        "log_level": "DEBUG",
        "otlp_traces_endpoints": ["http://otel:7777", "notused"],
        "admin_port": 8001,
    }

    translated_config = translate_dbos_config_to_config_file(config)

    assert translated_config["name"] == "test-app"
    assert translated_config["database"]["hostname"] == "localhost"
    assert translated_config["database"]["port"] == 5432
    assert translated_config["database"]["username"] == "user"
    assert translated_config["database"]["password"] == "password"
    assert translated_config["database"]["ssl"] == True
    assert translated_config["database"]["ssl_ca"] == "ca.pem"
    assert translated_config["database"]["connectionTimeoutMillis"] == 10000
    assert translated_config["database"]["app_db_name"] == "dbname"
    assert translated_config["database"]["sys_db_name"] == "sysdb"
    assert "database_url" not in translated_config
    assert translated_config["telemetry"]["logs"]["logLevel"] == "DEBUG"
    assert (
        translated_config["telemetry"]["OTLPExporter"]["tracesEndpoint"]
        == "http://otel:7777"
    )
    assert "logsEndpoint" not in translated_config["telemetry"]["OTLPExporter"]
    assert translated_config["runtimeConfig"]["admin_port"] == 8001
    assert "start" not in translated_config["runtimeConfig"]
    assert "setup" not in translated_config["runtimeConfig"]
    assert "env" not in translated_config


def test_translate_dbosconfig_minimal_input():
    config: DBOSConfig = {
        "name": "test-app",
    }
    translated_config = translate_dbos_config_to_config_file(config)

    assert translated_config["name"] == "test-app"
    assert translated_config["telemetry"]["logs"]["logLevel"] == "INFO"
    assert "database" not in translated_config
    assert "env" not in translated_config
    assert "runtimeConfig" not in translated_config


def test_translate_dbosconfig_just_sys_db_name():
    config: DBOSConfig = {
        "name": "test-app",
        "sys_db_name": "sysdb",
    }
    translated_config = translate_dbos_config_to_config_file(config)

    assert translated_config["name"] == "test-app"
    assert translated_config["telemetry"]["logs"]["logLevel"] == "INFO"
    assert translated_config["database"]["sys_db_name"] == "sysdb"
    assert "env" not in translated_config
    assert "runtimeConfig" not in translated_config


def test_translate_empty_otlp_traces_endpoints():
    # Give an empty OTLP traces endpoint list
    config: DBOSConfig = {
        "name": "test-app",
        "otlp_traces_endpoints": [],
    }
    translated_config = translate_dbos_config_to_config_file(config)
    assert "OTLPExporter" not in translated_config["telemetry"]
    assert translated_config["telemetry"]["logs"]["logLevel"] == "INFO"


def test_translate_ignores_otlp_traces_not_list():
    # Give an empty OTLP traces endpoint list
    config: DBOSConfig = {
        "name": "test-app",
        "otlp_traces_endpoints": "http://otel:7777",
    }
    translated_config = translate_dbos_config_to_config_file(config)
    assert translated_config["name"] == "test-app"
    assert "OTLPExporter" not in translated_config["telemetry"]


def test_translate_missing_name():
    with pytest.raises(DBOSInitializationError) as exc_info:
        translate_dbos_config_to_config_file({})
    assert (
        "Error initializing DBOS Transact: Configuration must specify an application name"
        in str(exc_info.value)
    )


####################
# CONFIG OVERWRITE
####################


def test_overwrite_config(mocker):
    # Setup a typical dbos-config.yaml file
    mock_config = """
    name: "stock-prices"
    language: "python"
    database:
        hostname: "hostname"
        port: 1234
        username: dbosadmin
        password: pwd
        app_db_name: appdbname
        sys_db_name: sysdbname
        ssl: true
        ssl_ca: cert.pem
        migrate:
            - alembic upgrade head
    telemetry:
        logs:
            logLevel: INFO
        OTLPExporter:
            logsEndpoint: thelogsendpoint
            tracesEndpoint:  thetracesendpoint
    env:
        KEY: "VALUE"
    runtimeConfig:
        start:
            - "a start command"
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    provided_config: ConfigFile = {
        "name": "test-app",
        "database": {
            "hostname": "localhost",
            "port": 5432,
            "username": "postgres",
            "password": "dbos",
            "app_db_name": "dbostestpy",
            "sys_db_name": "sysdb",
            "connectionTimeoutMillis": 10000,
        },
        "telemetry": {
            "OTLPExporter": {
                "tracesEndpoint": "a",
                "logsEndpoint": "b",
            },
            "logs": {
                "logLevel": "DEBUG",
            },
        },
        "runtimeConfig": {
            "admin_port": 8001,
        },
        "env": {
            "FOO": "BAR",
        },
    }
    config = overwrite_config(provided_config)

    assert config["name"] == "stock-prices"
    assert config["database"]["hostname"] == "hostname"
    assert config["database"]["port"] == 1234
    assert config["database"]["username"] == "dbosadmin"
    assert config["database"]["password"] == "pwd"
    assert config["database"]["app_db_name"] == "appdbname"
    assert config["database"]["sys_db_name"] == "sysdbname"
    assert config["database"]["ssl"] == True
    assert config["database"]["ssl_ca"] == "cert.pem"
    assert config["database"]["connectionTimeoutMillis"] == 10000
    assert config["telemetry"]["logs"]["logLevel"] == "DEBUG"
    assert config["telemetry"]["OTLPExporter"]["tracesEndpoint"] == "thetracesendpoint"
    assert config["telemetry"]["OTLPExporter"]["logsEndpoint"] == "thelogsendpoint"
    assert "admin_port" not in config["runtimeConfig"]
    assert "env" not in config


def test_overwrite_config_minimal(mocker):
    mock_config = """
    name: "stock-prices"
    language: "python"
    database:
        hostname: "hostname"
        port: 1234
        username: dbosadmin
        password: pwd
        app_db_name: appdbname
        sys_db_name: sysdbname
        ssl: true
        ssl_ca: cert.pem
        migrate:
            - alembic upgrade head
    telemetry:
        OTLPExporter:
            logsEndpoint: thelogsendpoint
            tracesEndpoint:  thetracesendpoint
    runtimeConfig:
        start:
            - "a start command"
        setup:
            - "echo 'hello'"
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    provided_config: ConfigFile = {
        "name": "test-app",
    }
    config = overwrite_config(provided_config)

    assert config["name"] == "stock-prices"
    assert config["database"]["hostname"] == "hostname"
    assert config["database"]["port"] == 1234
    assert config["database"]["username"] == "dbosadmin"
    assert config["database"]["password"] == "pwd"
    assert config["database"]["app_db_name"] == "appdbname"
    assert config["database"]["sys_db_name"] == "sysdbname"
    assert config["database"]["ssl"] == True
    assert config["database"]["ssl_ca"] == "cert.pem"
    assert config["telemetry"]["OTLPExporter"]["tracesEndpoint"] == "thetracesendpoint"
    assert config["telemetry"]["OTLPExporter"]["logsEndpoint"] == "thelogsendpoint"
    assert "runtimeConfig" not in config
    assert "env" not in config


def test_overwrite_config_has_telemetry(mocker):
    mock_config = """
    name: "stock-prices"
    language: "python"
    database:
        hostname: "hostname"
        port: 1234
        username: dbosadmin
        password: pwd
        app_db_name: appdbname
        sys_db_name: sysdbname
        ssl: true
        ssl_ca: cert.pem
        migrate:
            - alembic upgrade head
    telemetry:
        OTLPExporter:
            logsEndpoint: thelogsendpoint
            tracesEndpoint:  thetracesendpoint
    runtimeConfig:
        start:
            - "a start command"
        setup:
            - "echo 'hello'"
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    provided_config: ConfigFile = {
        "name": "test-app",
        "database": {
            "hostname": "localhost",
            "port": 5432,
            "username": "postgres",
            "password": "dbos",
            "app_db_name": "dbostestpy",
        },
        "telemetry": {"logs": {"logLevel": "DEBUG"}},
    }
    config = overwrite_config(provided_config)

    assert config["name"] == "stock-prices"
    assert config["database"]["hostname"] == "hostname"
    assert config["database"]["port"] == 1234
    assert config["database"]["username"] == "dbosadmin"
    assert config["database"]["password"] == "pwd"
    assert config["database"]["app_db_name"] == "appdbname"
    assert config["database"]["sys_db_name"] == "sysdbname"
    assert config["database"]["ssl"] == True
    assert config["database"]["ssl_ca"] == "cert.pem"
    assert config["telemetry"]["OTLPExporter"]["tracesEndpoint"] == "thetracesendpoint"
    assert config["telemetry"]["OTLPExporter"]["logsEndpoint"] == "thelogsendpoint"
    assert config["telemetry"]["logs"]["logLevel"] == "DEBUG"
    assert "runtimeConfig" not in config
    assert "env" not in config


# Not expected in practice, but exercise the code path
def test_overwrite_config_no_telemetry_in_file(mocker):
    mock_config = """
    name: "stock-prices"
    language: "python"
    database:
        hostname: "hostname"
        port: 1234
        username: dbosadmin
        password: pwd
        app_db_name: appdbname
        sys_db_name: sysdbname
        ssl: true
        ssl_ca: cert.pem
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    provided_config: ConfigFile = {
        "name": "test-app",
        "database": {
            "hostname": "localhost",
            "port": 5432,
            "username": "postgres",
            "password": "dbos",
            "app_db_name": "dbostestpy",
        },
        "telemetry": {"logs": {"logLevel": "DEBUG"}},
    }

    config = overwrite_config(provided_config)
    # Test that telemetry from provided_config is preserved
    assert config["telemetry"]["logs"]["logLevel"] == "DEBUG"
    assert config["telemetry"]["OTLPExporter"] == {}


# Not expected in practice, but exercise the code path
def test_overwrite_config_no_otlp_in_file(mocker):
    mock_config = """
    name: "stock-prices"
    language: "python"
    database:
        hostname: "hostname"
        port: 1234
        username: dbosadmin
        password: pwd
        app_db_name: appdbname
        sys_db_name: sysdbname
        ssl: true
        ssl_ca: cert.pem
    telemetry:
        logs:
            logLevel: INFO
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    provided_config: ConfigFile = {
        "name": "test-app",
        "database": {
            "hostname": "localhost",
            "port": 5432,
            "username": "postgres",
            "password": "dbos",
            "app_db_name": "dbostestpy",
        },
        "telemetry": {
            "OTLPExporter": {
                "tracesEndpoint": "original-trace",
                "logsEndpoint": "original-log",
            }
        },
    }

    config = overwrite_config(provided_config)
    # Test that OTLPExporter from provided_config is preserved
    assert config["telemetry"]["OTLPExporter"]["tracesEndpoint"] == "original-trace"
    assert config["telemetry"]["OTLPExporter"]["logsEndpoint"] == "original-log"
    assert "logs" not in config["telemetry"]


# Not expected in practice, but exercise the code path
def test_overwrite_config_with_provided_database_url(mocker):
    mock_config = """
    name: "stock-prices"
    language: "python"
    database:
        hostname: "hostname"
        port: 1234
        username: dbosadmin
        password: pwd
        app_db_name: appdbname
        sys_db_name: sysdbname
        ssl: true
        ssl_ca: cert.pem
        migrate:
            - alembic upgrade head
    telemetry:
        OTLPExporter:
            logsEndpoint: thelogsendpoint
            tracesEndpoint:  thetracesendpoint
    runtimeConfig:
        start:
            - "a start command"
        setup:
            - "echo 'hello'"
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )

    provided_config: ConfigFile = {
        "name": "test-app",
        "database_url": "ignored",
    }
    config = overwrite_config(provided_config)

    assert config["name"] == "stock-prices"
    assert config["database"]["hostname"] == "hostname"
    assert config["database"]["port"] == 1234
    assert config["database"]["username"] == "dbosadmin"
    assert config["database"]["password"] == "pwd"
    assert config["database"]["app_db_name"] == "appdbname"
    assert config["database"]["sys_db_name"] == "sysdbname"
    assert config["database"]["ssl"] == True
    assert config["database"]["ssl_ca"] == "cert.pem"
    assert config["telemetry"]["OTLPExporter"]["tracesEndpoint"] == "thetracesendpoint"
    assert config["telemetry"]["OTLPExporter"]["logsEndpoint"] == "thelogsendpoint"
    assert "runtimeConfig" not in config
    assert "env" not in config


####################
# PROVIDED CONFIGS vs CONFIG FILE
####################


def test_no_discrepancy(mocker):
    mock_config = """
    name: "stock-prices" \
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )
    check_config_consistency(name="stock-prices")


def test_name_does_no_match(mocker):
    mock_config = """
    name: "stock-prices" \
    """
    mocker.patch(
        "builtins.open", side_effect=generate_mock_open("dbos-config.yaml", mock_config)
    )
    with pytest.raises(DBOSInitializationError) as exc_info:
        check_config_consistency(name="stock-prices-wrong")
    assert (
        "Provided app name 'stock-prices-wrong' does not match the app name 'stock-prices' in dbos-config.yaml"
        in str(exc_info.value)
    )


def test_no_config_file():
    # Handles FileNotFoundError
    check_config_consistency(name="stock-prices")



import os
from unittest.mock import mock_open, patch

import pytest
import yaml

from dbos._db_wizard import db_wizard
from dbos._dbos_config import ConfigFile
from dbos._error import DBOSInitializationError


class TestDbWizardIntegration:
    """Integration test suite for the db_wizard function with a real PostgreSQL database."""

    @pytest.fixture
    def standard_config(self) -> ConfigFile:
        """Standard configuration with default database settings."""
        return {
            "name": "test-app",
            "database": {
                "hostname": "localhost",
                "port": 5432,
                "username": "postgres",
                "password": os.environ.get("PGPASSWORD", "dbos"),
                "app_db_name": "postgres",
            },
        }

    @pytest.fixture
    def non_default_config(self) -> ConfigFile:
        """Configuration with non-default database settings."""
        return {
            "name": "test-app",
            "database": {
                "hostname": "non-existent-host",
                "port": 5433,
                "username": "non-existent-user",
                "password": "wrong-password",
                "app_db_name": "postgres",
            },
        }

    @pytest.fixture
    def wrong_password_config(self) -> ConfigFile:
        """Configuration with correct host/port but wrong password."""
        return {
            "name": "test-app",
            "database": {
                "hostname": "localhost",
                "port": 5432,
                "username": "postgres",
                "password": "wrong",
                "app_db_name": "postgres",
            },
        }

    def test_successful_connection_postgres_db(
        self, standard_config: ConfigFile
    ) -> None:
        """Test when connection to postgres database is successful."""
        result = db_wizard(standard_config)
        assert result == standard_config

    '''
    Unfortunately psycopg gets a timeout error, even w/ the wrong password...
    def test_wrong_password(self, wrong_password_config):
        """Test with wrong password."""
        with pytest.raises(DBOSInitializationError) as exc_info:
            db_wizard(wrong_password_config, "config.yaml")
        assert "password authentication failed" in str(exc_info.value)
    '''

    def test_non_default_config_settings(self, non_default_config: ConfigFile) -> None:
        """Test with non-default database configuration."""
        mock_file_content = yaml.dump(
            {"name": "test-app"}
        )  # no db config in config file
        with patch("builtins.open", mock_open(read_data=mock_file_content)):
            # Should raise error because config has non-default database settings
            with pytest.raises(DBOSInitializationError) as exc_info:
                db_wizard(non_default_config)

            # Verify error message
            assert "Could not connect to the database" in str(exc_info.value)



import json
import os
import re
import sys
from importlib import resources
from typing import Any, Dict, List, Optional, TypedDict, Union, cast

if sys.version_info < (3, 10):
    from typing_extensions import TypeGuard
else:
    from typing import TypeGuard

import yaml
from jsonschema import ValidationError, validate
from rich import print
from sqlalchemy import URL, make_url

from ._db_wizard import db_wizard, load_db_connection
from ._error import DBOSInitializationError
from ._logger import dbos_logger

DBOS_CONFIG_PATH = "dbos-config.yaml"


class DBOSConfig(TypedDict):
    """
    Data structure containing the DBOS library configuration.

    Attributes:
        name (str): Application name
        database_url (str): Database connection string
        sys_db_name (str): System database name
        log_level (str): Log level
        otlp_traces_endpoints: List[str]: OTLP traces endpoints
    """

    name: str
    database_url: Optional[str]
    sys_db_name: Optional[str]
    log_level: Optional[str]
    otlp_traces_endpoints: Optional[List[str]]
    admin_port: Optional[int]


class RuntimeConfig(TypedDict, total=False):
    start: List[str]
    setup: Optional[List[str]]
    admin_port: Optional[int]


class DatabaseConfig(TypedDict, total=False):
    hostname: str
    port: int
    username: str
    password: str
    connectionTimeoutMillis: Optional[int]
    app_db_name: str
    sys_db_name: Optional[str]
    ssl: Optional[bool]
    ssl_ca: Optional[str]
    local_suffix: Optional[bool]
    migrate: Optional[List[str]]
    rollback: Optional[List[str]]


def parse_database_url_to_dbconfig(database_url: str) -> DatabaseConfig:
    db_url = make_url(database_url)
    db_config = {
        "hostname": db_url.host,
        "port": db_url.port or 5432,
        "username": db_url.username,
        "password": db_url.password,
        "app_db_name": db_url.database,
    }
    for key, value in db_url.query.items():
        str_value = value[0] if isinstance(value, tuple) else value
        if key == "connect_timeout":
            db_config["connectionTimeoutMillis"] = int(str_value) * 1000
        elif key == "sslmode":
            db_config["ssl"] = str_value == "require"
        elif key == "sslrootcert":
            db_config["ssl_ca"] = str_value
    return cast(DatabaseConfig, db_config)


class OTLPExporterConfig(TypedDict, total=False):
    logsEndpoint: Optional[str]
    tracesEndpoint: Optional[str]


class LoggerConfig(TypedDict, total=False):
    logLevel: Optional[str]


class TelemetryConfig(TypedDict, total=False):
    logs: Optional[LoggerConfig]
    OTLPExporter: Optional[OTLPExporterConfig]


class ConfigFile(TypedDict, total=False):
    """
    Data structure containing the DBOS Configuration.

    This configuration data is typically loaded from `dbos-config.yaml`.
    See `https://docs.dbos.dev/api-reference/configuration`_

    Attributes:
        name (str): Application name
        runtimeConfig (RuntimeConfig): Configuration for request serving
        database (DatabaseConfig): Configuration for the application and system databases
        database_url (str): Database connection string
        telemetry (TelemetryConfig): Configuration for tracing / logging
        env (Dict[str,str]): Environment varialbes
        application (Dict[str, Any]): Application-specific configuration section

    """

    name: str
    runtimeConfig: RuntimeConfig
    database: DatabaseConfig
    database_url: Optional[str]
    telemetry: Optional[TelemetryConfig]
    env: Dict[str, str]


def is_dbos_configfile(data: Union[ConfigFile, DBOSConfig]) -> TypeGuard[DBOSConfig]:
    """
    Type guard to check if the provided data is a DBOSConfig.

    Args:
        data: The configuration object to check

    Returns:
        True if the data is a DBOSConfig, False otherwise
    """
    return (
        isinstance(data, dict)
        and "name" in data
        and (
            "runtimeConfig" in data
            or "database" in data
            or "env" in data
            or "telemetry" in data
        )
    )


def translate_dbos_config_to_config_file(config: DBOSConfig) -> ConfigFile:
    if "name" not in config:
        raise DBOSInitializationError(f"Configuration must specify an application name")

    translated_config: ConfigFile = {
        "name": config["name"],
    }

    # Database config
    db_config: DatabaseConfig = {}
    database_url = config.get("database_url")
    if database_url:
        db_config = parse_database_url_to_dbconfig(database_url)
    if "sys_db_name" in config:
        db_config["sys_db_name"] = config.get("sys_db_name")
    if db_config:
        translated_config["database"] = db_config

    # Admin port
    if "admin_port" in config:
        translated_config["runtimeConfig"] = {"admin_port": config["admin_port"]}

    # Telemetry config
    telemetry = {}
    # Add OTLPExporter if traces endpoints exist
    otlp_trace_endpoints = config.get("otlp_traces_endpoints")
    if isinstance(otlp_trace_endpoints, list) and len(otlp_trace_endpoints) > 0:
        telemetry["OTLPExporter"] = {"tracesEndpoint": otlp_trace_endpoints[0]}
    # Default to INFO -- the logging seems to default to WARN otherwise.
    log_level = config.get("log_level", "INFO")
    if log_level:
        telemetry["logs"] = {"logLevel": log_level}
    if telemetry:
        translated_config["telemetry"] = cast(TelemetryConfig, telemetry)

    return translated_config


def _substitute_env_vars(content: str) -> str:
    regex = r"\$\{([^}]+)\}"  # Regex to match ${VAR_NAME} style placeholders

    def replace_func(match: re.Match[str]) -> str:
        var_name = match.group(1)
        value = os.environ.get(
            var_name, ""
        )  # If the env variable is not set, return an empty string
        if value == "":
            dbos_logger.warning(
                f"Variable {var_name} would be substituted from the process environment into dbos-config.yaml, but is not defined"
            )
        return value

    return re.sub(regex, replace_func, content)


def get_dbos_database_url(config_file_path: str = DBOS_CONFIG_PATH) -> str:
    """
    Retrieve application database URL from configuration `.yaml` file.

    Loads the DBOS `ConfigFile` from the specified path (typically `dbos-config.yaml`),
        and returns the database URL for the application database.

    Args:
        config_file_path (str): The path to the yaml configuration file.

    Returns:
        str: Database URL for the application database

    """
    dbos_config = process_config(data=load_config(config_file_path))
    db_url = URL.create(
        "postgresql+psycopg",
        username=dbos_config["database"]["username"],
        password=dbos_config["database"]["password"],
        host=dbos_config["database"]["hostname"],
        port=dbos_config["database"]["port"],
        database=dbos_config["database"]["app_db_name"],
    )
    return db_url.render_as_string(hide_password=False)


def load_config(
    config_file_path: str = DBOS_CONFIG_PATH,
    *,
    run_process_config: bool = True,
    use_db_wizard: bool = True,
    silent: bool = False,
) -> ConfigFile:
    """
    Load the DBOS `ConfigFile` from the specified path (typically `dbos-config.yaml`).

    The configuration is also validated against the configuration file schema.

    Args:
        config_file_path (str): The path to the yaml configuration file.

    Returns:
        ConfigFile: The loaded configuration

    """

    with open(config_file_path, "r") as file:
        content = file.read()
        substituted_content = _substitute_env_vars(content)
        data = yaml.safe_load(substituted_content)

    if not isinstance(data, dict):
        raise DBOSInitializationError(
            f"dbos-config.yaml must contain a dictionary, not {type(data)}"
        )
    data = cast(Dict[str, Any], data)

    # Load the JSON schema relative to the package root
    schema_file = resources.files("dbos").joinpath("dbos-config.schema.json")
    with schema_file.open("r") as f:
        schema = json.load(f)

    # Validate the data against the schema
    try:
        validate(instance=data, schema=schema)
    except ValidationError as e:
        raise DBOSInitializationError(f"Validation error: {e}")

    data = cast(ConfigFile, data)
    if run_process_config:
        data = process_config(data=data, use_db_wizard=use_db_wizard, silent=silent)
    return data  # type: ignore


def process_config(
    *,
    use_db_wizard: bool = True,
    data: ConfigFile,
    silent: bool = False,
) -> ConfigFile:

    if "name" not in data:
        raise DBOSInitializationError(f"Configuration must specify an application name")

    if not _is_valid_app_name(data["name"]):
        raise DBOSInitializationError(
            f'Invalid app name {data["name"]}.  App names must be between 3 and 30 characters long and contain only lowercase letters, numbers, dashes, and underscores.'
        )

    if "database" not in data:
        data["database"] = {}

    # database_url takes precedence over database config, but we need to preserve rollback and migrate if they exist
    migrate = data["database"].get("migrate", False)
    rollback = data["database"].get("rollback", False)
    local_suffix = data["database"].get("local_suffix", False)
    if data.get("database_url"):
        dbconfig = parse_database_url_to_dbconfig(cast(str, data["database_url"]))
        if migrate:
            dbconfig["migrate"] = cast(List[str], migrate)
        if rollback:
            dbconfig["rollback"] = cast(List[str], rollback)
        if local_suffix:
            dbconfig["local_suffix"] = cast(bool, local_suffix)
        data["database"] = dbconfig

    if "app_db_name" not in data["database"] or not (data["database"]["app_db_name"]):
        data["database"]["app_db_name"] = _app_name_to_db_name(data["name"])

    # Load the DB connection file. Use its values for missing fields from dbos-config.yaml. Use defaults otherwise.
    db_connection = load_db_connection()
    if not silent:
        if os.getenv("DBOS_DBHOST"):
            print(
                "[bold blue]Loading database connection parameters from debug environment variables[/bold blue]"
            )
        elif data["database"].get("hostname"):
            print(
                "[bold blue]Loading database connection parameters from dbos-config.yaml[/bold blue]"
            )
        elif db_connection.get("hostname"):
            print(
                "[bold blue]Loading database connection parameters from .dbos/db_connection[/bold blue]"
            )
        else:
            print(
                "[bold blue]Using default database connection parameters (localhost)[/bold blue]"
            )

    dbos_dbport: Optional[int] = None
    dbport_env = os.getenv("DBOS_DBPORT")
    if dbport_env:
        try:
            dbos_dbport = int(dbport_env)
        except ValueError:
            pass
    dbos_dblocalsuffix: Optional[bool] = None
    dblocalsuffix_env = os.getenv("DBOS_DBLOCALSUFFIX")
    if dblocalsuffix_env:
        try:
            dbos_dblocalsuffix = dblocalsuffix_env.casefold() == "true".casefold()
        except ValueError:
            pass

    data["database"]["hostname"] = (
        os.getenv("DBOS_DBHOST")
        or data["database"].get("hostname")
        or db_connection.get("hostname")
        or "localhost"
    )

    data["database"]["port"] = (
        dbos_dbport or data["database"].get("port") or db_connection.get("port") or 5432
    )
    data["database"]["username"] = (
        os.getenv("DBOS_DBUSER")
        or data["database"].get("username")
        or db_connection.get("username")
        or "postgres"
    )
    data["database"]["password"] = (
        os.getenv("DBOS_DBPASSWORD")
        or data["database"].get("password")
        or db_connection.get("password")
        or os.environ.get("PGPASSWORD")
        or "dbos"
    )

    local_suffix = False
    dbcon_local_suffix = db_connection.get("local_suffix")
    if dbcon_local_suffix is not None:
        local_suffix = dbcon_local_suffix
    db_local_suffix = data["database"].get("local_suffix")
    if db_local_suffix is not None:
        local_suffix = db_local_suffix
    if dbos_dblocalsuffix is not None:
        local_suffix = dbos_dblocalsuffix
    data["database"]["local_suffix"] = local_suffix

    # Check the connectivity to the database and make sure it's properly configured
    # Note, never use db wizard if the DBOS is running in debug mode (i.e. DBOS_DEBUG_WORKFLOW_ID env var is set)
    debugWorkflowId = os.getenv("DBOS_DEBUG_WORKFLOW_ID")
    if use_db_wizard and debugWorkflowId is None:
        data = db_wizard(data)

    if "local_suffix" in data["database"] and data["database"]["local_suffix"]:
        data["database"]["app_db_name"] = f"{data['database']['app_db_name']}_local"

    # Return data as ConfigFile type
    return data


def _is_valid_app_name(name: str) -> bool:
    name_len = len(name)
    if name_len < 3 or name_len > 30:
        return False
    match = re.match("^[a-z0-9-_]+$", name)
    return True if match != None else False


def _app_name_to_db_name(app_name: str) -> str:
    name = app_name.replace("-", "_")
    return name if not name[0].isdigit() else f"_{name}"


def set_env_vars(config: ConfigFile) -> None:
    for env, value in config.get("env", {}).items():
        if value is not None:
            os.environ[env] = str(value)


def overwrite_config(provided_config: ConfigFile) -> ConfigFile:
    # Load the DBOS configuration file and force the use of:
    # 1. The database connection parameters (sub the file data to the provided config)
    # 2. OTLP traces endpoints (add the config data to the provided config)
    # 3. Use the application name from the file. This is a defensive measure to ensure the application name is whatever it was registered with in the cloud
    # 4. Remove admin_port is provided in code
    # 5. Remove env vars if provided in code
    # Optimistically assume that expected fields in config_from_file are present

    config_from_file = load_config(run_process_config=False)
    # Be defensive
    if config_from_file is None:
        return provided_config

    # Name
    provided_config["name"] = config_from_file["name"]

    # Database config. Note we disregard a potential database_url in config_from_file because it is not expected from DBOS Cloud
    if "database" not in provided_config:
        provided_config["database"] = {}
    provided_config["database"]["hostname"] = config_from_file["database"]["hostname"]
    provided_config["database"]["port"] = config_from_file["database"]["port"]
    provided_config["database"]["username"] = config_from_file["database"]["username"]
    provided_config["database"]["password"] = config_from_file["database"]["password"]
    provided_config["database"]["app_db_name"] = config_from_file["database"][
        "app_db_name"
    ]
    provided_config["database"]["sys_db_name"] = config_from_file["database"][
        "sys_db_name"
    ]
    provided_config["database"]["ssl"] = config_from_file["database"]["ssl"]
    provided_config["database"]["ssl_ca"] = config_from_file["database"]["ssl_ca"]

    # Telemetry config
    if "telemetry" not in provided_config or provided_config["telemetry"] is None:
        provided_config["telemetry"] = {
            "OTLPExporter": {},
        }
    elif "OTLPExporter" not in provided_config["telemetry"]:
        provided_config["telemetry"]["OTLPExporter"] = {}

    # This is a super messy from a typing perspective.
    # Some of ConfigFile keys are optional -- but in practice they'll always be present in hosted environments
    # So, for Mypy, we have to (1) check the keys are present in config_from_file and (2) cast telemetry/otlp_exporters to Dict[str, Any]
    # (2) is required because, even tho we resolved these keys earlier, mypy doesn't remember that
    if (
        config_from_file.get("telemetry")
        and config_from_file["telemetry"]
        and config_from_file["telemetry"].get("OTLPExporter")
    ):

        telemetry = cast(Dict[str, Any], provided_config["telemetry"])
        otlp_exporter = cast(Dict[str, Any], telemetry["OTLPExporter"])

        source_otlp = config_from_file["telemetry"]["OTLPExporter"]
        if source_otlp:
            tracesEndpoint = source_otlp.get("tracesEndpoint")
            if tracesEndpoint:
                otlp_exporter["tracesEndpoint"] = tracesEndpoint
            logsEndpoint = source_otlp.get("logsEndpoint")
            if logsEndpoint:
                otlp_exporter["logsEndpoint"] = logsEndpoint

    # Runtime config
    if (
        "runtimeConfig" in provided_config
        and "admin_port" in provided_config["runtimeConfig"]
    ):
        del provided_config["runtimeConfig"][
            "admin_port"
        ]  # Admin port is expected to be 3001 (the default in dbos/_admin_server.py::__init__ ) by DBOS Cloud

    # Env should be set from the hosting provider (e.g., DBOS Cloud)
    if "env" in provided_config:
        del provided_config["env"]

    return provided_config


def check_config_consistency(
    *,
    name: str,
    config_file_path: str = DBOS_CONFIG_PATH,
) -> None:
    # First load the config file and check whether it is present
    try:
        config = load_config(config_file_path)
    except FileNotFoundError:
        dbos_logger.debug(
            f"No configuration file {config_file_path} found. Skipping consistency check with provided config."
        )
        return
    except Exception as e:
        raise e

    # Check the name
    if name != config["name"]:
        raise DBOSInitializationError(
            f"Provided app name '{name}' does not match the app name '{config['name']}' in {config_file_path}."
        )
